__author__ = 'lovit'
__name__ = 'soylemma: Korean trained lemmatizer'
__version__ = '0.1.0'

from .lemmatizer import Lemmatizer
from .lemmatizer import analyze_morphology
from .lemmatizer import get_lemma_candidates
from .hangle import compose
from .hangle import decompose
from .hangle import is_hangle
from .trainer import extract_rule
from .trainer import extract_rules
from .trainer import load_word_morpheme_table
from .utils import installpath

# tagset
from .utils import ADJECTIVE
from .utils import VERB
from .utils import EOMI


from collections import defaultdict
from .hangle import decompose

is_jaum = lambda c: 'ㄱ' <= c <= 'ㅎ'
is_moum = lambda c: 'ㅏ' <= c <= 'ㅣ'

def parse(line):
    eojeol, morphtags, count = line.strip().split('\t')
    morphtags = [mt.rsplit('/', 1) for mt in morphtags.split(' + ')]
    count = int(count)
    return eojeol, morphtags, count

def _right_form(morph):
    """
    Arguments
    ---------
    morph : str
        Morpheme str

    Returns
    -------
    형태소의 첫 글자가 모음, 두번째 글자가 자음이면 False, 그 외에는 True
    eg)
        eojeol = '갔다가'
        morphemes = '가/VV + ㅏㅆ/EP + 다가/EC'

    Usage
    -----
        print(_right_form('ㅏㅆ)) # False
        print(_right_form('다가)) # True
    """

    if is_moum(morph) and len(morph) > 1 and is_jaum(morph):
        return False
    return True

def right_form(morphemes):
    for morph, _ in morphemes:
        if not _right_form(morph):
            return False
    return True

def load_word_morpheme_table(path):
    """
    Arguments
    ---------
    path : str
        Eojeol, Morpheme, Count table

        File example, 

            개봉된	개봉되/Verb + ㄴ/Eomi	17
            개봉될	개봉되/Verb + ㄹ/Eomi	7
            개봉인	개봉이/Adjective + ㄴ/Eomi	2
            ...

    Returns
    -------
    eojeol_to_morphemes : list of tuple

        For example,

        eojeol_to_morphemes = [
            ...
            ('개정하면서', [['개정하', 'Verb'], ['면서', 'Eomi']]),
            ('개정하여', [['개정하', 'Verb'], ['아', 'Eomi']]),
            ('개정하자는', [['개정하', 'Verb'], ['자는', 'Eomi']])
            ...
        ]
    """
    # Eojeol, Morpheme, Count
    eojeol_to_morphemes = {}
    with open(path, encoding='utf-8') as f:
        next(f)
        for line in f:
            eojeol, morphemes, count = parse(line)
            if not right_form(morphemes):
                continue
            if eojeol in eojeol_to_morphemes:
                continue
            eojeol_to_morphemes[eojeol] = morphemes
    eojeol_to_morphemes = list(eojeol_to_morphemes.items())
    return eojeol_to_morphemes

def extract_rule(eojeol, lw, lt, rw, rt):
    """
    Arguments
    ---------
    eojeol : str
        Eojeol
    lr : str
        Left-side morpheme
    lt : str
        Tag of left-side morpheme
    rw : str
        Right-side morpheme
    rt : str
        Tag of right-side morpheme

    Returns
    -------
    surface, canon : str, str
        If the eojeol is conjugated it return surface & canon tuple
        Else, it return None

    Usage
    -----
        >>> extract_rule('가까웠는데', '가깝', 'Adjective', '었는데', 'Eomi')
        $ ('까웠', ('깝', '었는'))

        >>> extract_rule('가까워지며', '가까워지', 'Verb', '며', 'Eomi')
        $ None
    """

    if not (lt == 'Adjective' or lt == 'Verb'):
        return
    surface = eojeol[len(lw)-1:len(lw)+1]
    if decompose(surface[0])[0] != decompose(lw[-1])[0]:
        return
    if lw + rw == eojeol:
        return
    if len(lw) + len(rw) == len(eojeol):
        canon = (lw[-1], rw[0])
    elif len(lw) + len(rw) > len(eojeol):
        canon = (lw[-1], rw[:2])
    elif len(lw) + len(rw) + 1 == len(eojeol):
        surface = eojeol[len(lw)-1:len(lw)+2]
        canon = (lw[-1], rw[0])
    else:
        raise ValueError('처리 불가. eojeol={}, {}/{} + {}/{}'.format(eojeol, lw, lt, rw, rt))

    # post-processing
    if len(surface) == 2 and len(canon[0]) == 1 and len(canon[1]) == 1:
        surf_cho = (decompose(surface[0])[0], decompose(surface[1])[0])
        canon_cho = (decompose(canon[0][0])[0], decompose(canon[1][0])[0])
        if not (surf_cho[0] == canon_cho[0] and surf_cho[1] == canon_cho[1]):
            return

    return surface, canon

def extract_rules(eojeol_lr_array):
    """
    Arguments
    ---------
    eojeol_lr_array : nested list

        [
            (Eojeol, ((lw, lt), (rw, rt))),
            (Eojeol, ((lw, lt), (rw, rt))),
            ...
        ]
        All Eojeol, lw, lt, rw, rt is str type

    Returns
    -------
    rules : dict of set
        Lemmatizing rule
        rules = {
            '했던': {'하았던'},
            '인': {'이ㄴ'},
            ...
        }

    Usage
    -----
        eojeol_lr_array = [
            ('가당하시냐고', [['가당하', 'Adjective'], ['시냐고', 'Eomi']])
            ('가당하지', [['가당하', 'Adjective'], ['지', 'Eomi']])
            ('가당한', [['가당하', 'Adjective'], ['ㄴ', 'Eomi']])
            ('가닿는', [['가닿', 'Verb'], ['는', 'Eomi']])
            ('가닿는다는', [['가닿', 'Verb'], ['는다는', 'Eomi']])
            ...
        ]

        rules = extract_rules(eojeol_lr_array)
    """

    rules = defaultdict(lambda: set())
    for eojeol, ((lw, lt), (rw, rt)) in eojeol_lr_array:
        try:
            rule = extract_rule(eojeol, lw, lt, rw, rt)
            if rule is None:
                continue
            surface, canon = rule
            rules[surface].add(canon)
        except Exception as e:
            print(e)
            print(eojeol, ((lw, lt), (rw, rt)), end='\n\n')
    return dict(rules)


import re


hangle_pattern = re.compile('[가-힣]+')

def is_hangle(word):
    """
    Arguments
    ---------
    word : str

    Returns
    -------
    It returns True if all characters in the word are Hangle
    Else It returns False
    """

    match = hangle_pattern.match(word)
    if not match:
        return False
    span = match.span()
    return (span[1] - span[0]) == len(word)

def compose(cho, jung, jong):
    """
    Arguments
    ---------
    cho : str
        Chosung, length is 1
    jung : str
        Jungsung, length is 1
    jong : str
        Jongsung, length is 1

    Returns
    -------
    Composed hangle : str
    """

    cho_ = cho_to_idx.get(cho, -1)
    jung_ = jung_to_idx.get(jung, -1)
    jong_ = jong_to_idx.get(jong, -1)
    if (cho_ < 0) or (jung_ < 0) or (jong_ < 0):
        return None
    return chr(kor_begin + cho_base * cho_ + jung_base * jung_ + jong_)

def decompose(input, ensure_input=False):
    """
    Arguments
    ---------
    input : str
        Character, length is 1
    ensure_input : Boolean
        If True, pass length and hangle check

    Returns
    -------
    (cho, jung, jong) : tuple of str or None
        If input is hangle.
        Else it return None
    """

    if not ensure_input:
        if len(input) > 1 or not is_hangle(input):
            return None
    i = ord(input) - kor_begin
    cho  = i // cho_base
    jung = ( i - cho * cho_base ) // jung_base
    jong = ( i - cho * cho_base - jung * jung_base )
    return (chosungs[cho], jungsungs[jung], jongsungs[jong])

kor_begin = 44032
kor_end = 55203
cho_base = 588
jung_base = 28
jaum_begin = 12593
jaum_end = 12622
moum_begin = 12623
moum_end = 12643

# 19 characters
chosungs = [
    'ㄱ', 'ㄲ', 'ㄴ', 'ㄷ', 'ㄸ',
    'ㄹ', 'ㅁ', 'ㅂ', 'ㅃ', 'ㅅ',
    'ㅆ', 'ㅇ', 'ㅈ', 'ㅉ', 'ㅊ',
    'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ'
]

# 21 characters
jungsungs = [
    'ㅏ', 'ㅐ', 'ㅑ', 'ㅒ', 'ㅓ',
    'ㅔ', 'ㅕ', 'ㅖ', 'ㅗ', 'ㅘ',
    'ㅙ', 'ㅚ', 'ㅛ', 'ㅜ', 'ㅝ',
    'ㅞ', 'ㅟ', 'ㅠ', 'ㅡ', 'ㅢ',
    'ㅣ'
]

# 28 characters including white space
jongsungs = [
    ' ', 'ㄱ', 'ㄲ', 'ㄳ', 'ㄴ',
    'ㄵ', 'ㄶ', 'ㄷ', 'ㄹ', 'ㄺ',
    'ㄻ', 'ㄼ', 'ㄽ', 'ㄾ', 'ㄿ',
    'ㅀ', 'ㅁ', 'ㅂ', 'ㅄ', 'ㅅ',
    'ㅆ', 'ㅇ', 'ㅈ', 'ㅊ', 'ㅋ',
    'ㅌ', 'ㅍ', 'ㅎ'
]

# 30 characters
jaums = [
    'ㄱ', 'ㄲ', 'ㄳ', 'ㄴ', 'ㄵ',
    'ㄶ', 'ㄷ', 'ㄸ', 'ㄹ', 'ㄺ',
    'ㄻ', 'ㄼ', 'ㄽ', 'ㄾ', 'ㄿ',
    'ㅀ', 'ㅁ', 'ㅂ', 'ㅃ', 'ㅄ',
    'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅉ',
    'ㅊ', 'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ'
]

# 21 characters
moums = [
    'ㅏ', 'ㅐ', 'ㅑ', 'ㅒ', 'ㅓ',
    'ㅔ', 'ㅕ', 'ㅖ', 'ㅗ', 'ㅘ',
    'ㅙ', 'ㅚ', 'ㅛ', 'ㅜ', 'ㅝ',
    'ㅞ', 'ㅟ', 'ㅠ', 'ㅡ', 'ㅢ',
    'ㅣ'
]

cho_to_idx = {cho:idx for idx, cho in enumerate(chosungs)}
jung_to_idx = {jung:idx for idx, jung in enumerate(jungsungs)}
jong_to_idx = {jong:idx for idx, jong in enumerate(jongsungs)}


from collections import defaultdict
from .utils import installpath
from .utils import VERB, ADJECTIVE, EOMI


class Lemmatizer:
    """
    Korean trained lemmatizer class

    Arguments
    ---------
    verbs, adjectives, eomis : set of str
        Dictionary set
        If they are None, use trained dictionary.
    lemma_rules : dict
        Dictionary of lemmatization rules.
        For example,
            lemma_rules = {
                '했': {('하', '았')},
                '끔': {('끈', 'ㅁ'), ('끌', 'ㅁ')}
                '가우니': {('갑', '니')} # 차가우니까 -> 차갑 + 니까
                ...
            }
    dictionary_name : str
        Dictionary name.
        User can use their dictionary
        Dictionary file path is soylemma/dictionary/[dictionary_name]/
        Each dictionary must have four files
            |-- Adjectives.txt
            |-- Eomis.txt
            |-- Verbs.txt
            |-- rules.txt

    Usage
    -----

        >>> from soylemma import Lemmatizer

        >>> lemmatizer = Lemmatizer(dictionary_name='demo')
        >>> lemmatizer = Lemmatizer(dictionary_name='default')

        >>> lemmatizer.lemmatize('차가우니까')
        $ [('차갑다', 'Adjective')]

        >>> lemmatizer.analyze('차가우니까')
        $ [(('차갑', 'Adjective'), ('우니까', 'Eomi'))]

        >>> lemmatizer.conjugate('차갑', '우니까')
        $ ['차가우니까', '차갑우니까']

    """

    def __init__(self, verbs=None, adjectives=None,
        eomis=None, lemma_rules=None, dictionary_name='default'):

        verbs, adjectives, eomis = self._check_dictionary(
            verbs, adjectives, eomis, dictionary_name)

        lemma_rules, conjugate_rules = self._check_rules(
            lemma_rules, dictionary_name)

        self.verbs = verbs
        self.adjectives = adjectives
        self.eomis = eomis
        self.lemma_rules = lemma_rules
        self.conjugate_rules = conjugate_rules

    def _check_dictionary(self, verbs, adjectives, eomis, dictionary_name):
        """
        Arguments
        ---------
        verbs, adjectives, eomis : set of str
            Dictionary set
            If they are None, use trained dictionary.
            They are passed from __init__ function.

        Returns
        -------
        verbs, adjectives, eomis : set of str
            If each set is None, use trained dictionary with loading function.
        """

        morphs_set = [
            # morphs set, name
            (verbs, 'Verbs'),
            (adjectives, 'Adjectives'),
            (eomis, 'Eomis')
        ]
        morphs_set_ = []
        for morphs, tag in morphs_set:
            if morphs is None:
                morphs = self._load_dictionary(
                    '{}/soylemma/dictionary/{}/{}.txt'.format(
                        installpath, dictionary_name, tag))
            if not isinstance(morphs, set):
                morphs = set(morphs)
            morphs_set_.append(morphs)

        verbs, adjectives, eomis = morphs_set_
        return verbs, adjectives, eomis

    def _load_dictionary(self, path):
        """
        Arguments
        ---------
        path : str
            Dictionary file path

        Dictionary file can have information column such as word count
        For example,

            가 100
            먹 100
            시키 50

        However, it load only words, the first column in the file.

        Returns
        -------
        morphs : set of str
            Loaded dictionary
        """

        with open(path, encoding='utf-8') as f:
            morphs = {morph.split()[0] for morph in f}
        return morphs

    def _check_rules(self, lemma_rules, dictionary_name):
        """
        Arguments
        ---------
        lemma_rules : dict
            Dictionary of lemmatization rules.
            Passed from __init__ function

        Returns
        -------
        lemma_rules : dict
        conjugate_rules : dict
            Inverse mapper of lemma_rules
        """

        if lemma_rules is None:
            lemma_rules = self._load_rules(
                '{}/soylemma/dictionary/{}/rules.txt'.format(
                    installpath, dictionary_name))
        conjugate_rules = to_conjugate_rules(lemma_rules)
        return lemma_rules, conjugate_rules

    def _load_rules(self, path):
        """
        Arguments
        ---------
        path : str
            File path of rule table

        Rule table must have three column
        <surfacial form, canonical form of stem, canonical form of eomi>
        For example,

            했던 하 았던

        Returns
        -------
        lemma_rules : dict
        """

        with open(path, encoding='utf-8') as f:
            lines = [l.split() for l in f]
        lines = [(l[0], l[1], '아') if len(l) == 2 else l for l in lines]

        # 했던 -> (하, 았던)
        lemma_rules = defaultdict(lambda: set())
        for surf, stem, eomi in lines:
            lemma_rules[surf].add((stem, eomi))
        return dict(lemma_rules)

    def add_words(self, words, tag):
        """
        Arguments
        ---------
        words : collection of str
            Words
        tag : str
            Tag. choice from ['Adjective', 'Verb', 'Eomi']
        """

        # check words
        if isinstance(words, str):
            words = {words}

        if tag == ADJECTIVE:
            self.adjectives.update(words)
        elif tag == VERB:
            self.verbs.update(words)
        elif tag == EOMI:
            self.eomis.update(words)
        else:
            raise ValueError("You put wrong tag '{}'. Acceptable only ['Adjective', 'Verb', 'Eomi']".format(tag))

    def add_lemma_rules(self, rules):
        """
        Arguments
        ---------
        rules : lemma_rules
            Format example,

            lemma_rules = {
                '했': {('하', '았')},
                '끔': {('끈', 'ㅁ'), ('끌', 'ㅁ')}
                '가우니': {('갑', '니')} # 차가우니까 -> 차갑 + 니까
                ...
            }

        It first check input format, and update (lemma rules, conjugate rules) both
        """

        rules = check_rules(rules)
        self.lemma_rules = update_rules(self.lemma_rules, rules)

        supplements = to_conjugate_rules(rules)
        self.conjugate_rules = update_rules(self.conjugate_rules, supplements)

    def analyze(self, word, debug=False):
        """
        Arguments
        ---------
        word : str
            A word to perform morphological analysis
        debug : Boolean
            If True, verbose on

        Returns
        -------
        morphemes : list of tuple

        Usage
        -----
            >>> lemmatizer.analyze('차가우니까')
            $ [(('차갑', 'Adjective'), ('우니까', 'Eomi'))]
        """

        return analyze_morphology(
            word, self.verbs, self.adjectives,
            self.eomis, self.lemma_rules, debug)

    def lemmatize(self, word):
        """
        Arguments
        ---------
        word : str
            A word to recover canonical form (lemma)

        Returns
        -------
        morphemes : list of tuple

        Usage
        -----
            >>> lemmatizer.lemmatize('차가우니까')
            $ [('차갑다', 'Adjective')]
        """

        morphs = analyze_morphology(
            word, self.verbs, self.adjectives,
            self.eomis, self.lemma_rules)
        lemmas = [(stem[0]+'다', stem[1]) for stem, eomi in morphs]
        return lemmas

    def conjugate(self, stem, eomi):
        """
        Arguments
        ---------
        stem : str
        eomi : str

        Returns
        -------
        conjugated form : list of str

        Usage
        -----
            >>> lemmatizer.conjugate('차갑', '우니까')
            $ ['차가우니까', '차갑우니까']
        """

        return get_conjugate_candidates(stem, eomi, self.conjugate_rules)

def to_conjugate_rules(lemma_rules):
    # (하, 았) -> [했]
    conjugate_rules = defaultdict(lambda: set())
    for surf, canons in lemma_rules.items():
        for stem, eomi in canons:
            conjugate_rules[(stem, eomi)].add(surf)
    return dict(conjugate_rules)

def analyze_morphology(word, verbs, adjectives, eomis, lemma_rules, debug=False):
    """
    Arguments
    ---------
    word : str
        A word to analyze its morphology
    verbs : set of str
        Verb dictionary
    adjectives : set of str
        Adjective dictionary
    eomis : set of str
        Eomi dictionary
    lemma_rules : dict of tuple
        Lemmatization rules
    debug : Boolean
        If True, it prints all candidates

    Returns
    -------
    morphs : list of tuple
        For example,

            word = '파랬던'
            morphs = [(('파랗', 'Adjective'), ('았던', 'Eomi'))]

        Dictionary checked list of (stem, eomi)

    Function get_lemma_candidates returns set of (stem, eomi) candidates.
    This function checks whether the stem and eomi is known words using dictionaries.
    """

    morphs = set()
    for stem, eomi in get_lemma_candidates(word, lemma_rules, debug):
        if not (eomi in eomis):
            continue
        if stem in adjectives:
            morphs.add(((stem, ADJECTIVE), (eomi, EOMI)))
        if stem in verbs:
            morphs.add(((stem, VERB), (eomi, EOMI)))
    return list(morphs)

def get_lemma_candidates(word, rules, debug=False):
    """
    Arguments
    ---------
    word : str
        A word to analyze its morphology
    rules : dict of tuple
        Lemmatization rules

    Returns
    -------
    morphs : list of tuple
        All possible subword combination satisfying lemmatization rules


    용언이 활용되는 지점은 어간과 어미가 만나는 지점으로, 표현형 (surfacial form) 에서
    활용이 되는 지점의 길이에 따라 모든 경우를 확인한다.

    # 1 음절만 활용되는 경우
    - `했 = 하 + 았`
        - 시작했으니까 = 시작하 + 았으니까
    - `랬 = 랗 + 았`
        - 파랬던 = 파랗 + 았던

    # 2 음절만 활용되는 경우
    - `추운 = 춥 + 은`
        - 추운데 = 춥 + 은데
    - `했다 = 하 + 았다`
        - 시작했다 = 시작하 + 았다

    # 3 음절만 활용되는 경우
    - `가우니 = 갑 + 니`
        - 차가우니까 = 차갑 + 니까

    Debug mode 에서는 단어의 활용 지점과 단어의 어간, 어미 조합 후보를 출력한다.

        >>> lemmatizer = Lemmatizer(dictionary_name='demo')
        >>> lemmatizer.analyze('파랬다', debug=True)

        $ [DEBUG] word: 파랬다 = 파랗 + 았다, conjugation: 랬 = 랗 + 았
    """

    def debug_on(word, l, stem, eomi, r, conj):
        args = (word, l+stem, eomi+r, conj, stem, eomi)
        print('[DEBUG] word: {} = {} + {}, conjugation: {} = {} + {}'.format(*args))

    max_i = len(word) - 1
    candidates = []
    for i, c in enumerate(word):
        l = word[:i+1]
        r = word[i+1:]
        l_ = word[:i]
        if i < max_i:
            candidates.append((l, r))

        # 1 syllable conjugation
        for stem, eomi in rules.get(c, {}):
            for stem, eomi in rules.get(c, {}):
                candidates.append((l_ + stem, eomi + r))
                if debug:
                    debug_on(word, l_, stem, eomi, r, c)

        # 2 or 3 syllables conjugation
        for conj in {word[i:i+2], word[i:i+3]}:
            for stem, eomi in rules.get(conj, {}):
                candidates.append((l_ + stem, eomi + r[1:]))
                if debug:
                    debug_on(word, l_, stem, eomi, r[1:], conj)
    return candidates

def get_conjugate_candidates(stem, eomi, rules):
    stem_ = stem[:-1]
    eomi_ = eomi[1:]
    key = (stem[-1], eomi[0])
    candidates = ['{}{}{}'.format(stem_, surface, eomi_) for surface in rules.get(key, {})]
    if len(eomi) >= 2:
        key = (stem[-1], eomi[:2])
        eomi_ = eomi[2:]
        candidates += ['{}{}{}'.format(stem_, surface, eomi_) for surface in rules.get(key, {})]
    candidates.append(stem + eomi)
    return candidates

def check_rules(rules):
    def type_error():
        raise ValueError("Wrong format inserted rules. rules={surface:{(stem, eomi), (stem, eomi), ...}}")

    rules_ = {}
    try:
        for surface, canons in rules.items():
            if isinstance(canons, str) or not isinstance(surface, str):
                type_error()
            canons_ = set()
            for canon in canons:
                if not (len(canon) == 2 and isinstance(canon[0], str) and isinstance(canon[1], str)):
                    type_error()
                canons_.add((canon[0], canon[1]))
            rules_[surface] = canons_
        return rules_
    except Exception as e:
        raise ValueError(str(e))

def update_rules(base, supplement):
    for surface, supple_set in supplement.items():
        base_set = base.get(surface, set())
        base_set.update(supple_set)
        base[surface] = base_set
    return base



import soylemma
import setuptools
from setuptools import setup, find_packages


with open('README.md', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="soylemma",
    version=soylemma.__version__,
    author=soylemma.__author__,
    author_email='soy.lovit@gmail.com',
    description="Trained Korean Lemmatizer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/lovit/korean_lemmatizer',
    packages=setuptools.find_packages(),
    package_data={
        'soylemma':[
            'dictionary/default/*'
        ]
    },
    keywords = [
        'korean-nlp',
        'nlp',
        'lemmatizer',
    ],
    classifiers=(
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ),
)
